'''
Action interscanners for sagator, version 0.6.0

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import re,os,time
from match import match_any

__all__=['quarantine','drop','deliver','deliver_to']

#####################################################################
### QUARANTINE class

class quarantine(match_any):
  '''
  quarantine() - quarantine into any directory
  
  usage: quarantine('qdir','dont quarantine regular expression',scanners)

  where: 'qdir' is a string, which defines a qatantine directory,
                directory, in which viruses/spams are stored.
                This string can contain %X substrings. These strings
                are represented as defined in strftime. For more information
                see "man strftime". If this directory not exitst,
                it's created.
         'dont quarantine regular expression' is a regular expression,
                which defines virus names, which can't be quarantined

  Example: quarantine('/tmp/quarantine/%Y%m', a_scanner())
             - every month will be stored in separate directory

  This scanner can be used to quarantine an virus/spam.
  qdir is a directory, in which viruses are stored.
  Following regular expression can be used to disable quarantining
  of some viruses or spams.
  '''
  name='quarantine()'
  def __init__(self,qdir,dontq,*scanners):
      self.QDIR=qdir
      self.DONTQ=dontq
      match_any.__init__(self,scanners)
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if globals.scan_only:
        return level,detected,virlist
      if is_infected(level,detected):
        if (self.QDIR!='') & (self.DONTQ!='') \
            & (re.search(self.DONTQ,detected,re.IGNORECASE)!=None):
          qfname='Not quarantined yet.'
        else:
          fext=time.strftime("-%Y%m%d-%H%M%S-",time.localtime(time.time()))
          fpath=time.strftime(self.QDIR)
          # try to make this directory
          try:
            os.makedirs(fpath)
          except OSError,(ec,es):
            if ec!=17: # file exists
              raise
          num=0
          while 1:
            qfname=fpath+'/qm'+fext+str(num)
            num+=1
            try:
              fd=safe.osopen(qfname,os.O_WRONLY|os.O_CREAT|os.O_EXCL,0660)
              f=os.fdopen(fd,'w',BUFSIZE)
              f.write(mail.addr)
              f.write(mail.comm)
              f.write(mail.xheader) # added headers
              f.write(mail.data) # original data, not modified
              f.write(".\r\nQUIT\r\n")
              f.close()
              debug.echo(3,"Quarantined as: "+qfname)
              break
            except OSError,(ec,es):
              if ec==17: # already exists
                continue
              else:
                debug.echo(2,"ERROR: quarantine(): ",es)
                #return level,detected,virlist
                raise ScannerError,"qarantine(): OSError: "+es
        globals.QFNAME=qfname
      return level,detected,virlist

#####################################################################
### drop and deliver classes

class drop(match_any):
  '''
  drop() - a class to DROP viruses/spams, by default they are REJECTED

  usage: drop(drop_pattern,scanners)
  
  This scanner can be used to drop some viruses, like Klez.
  Klez sends fake sender address and this virus can't be sent
  back to sender, because sender is faked.
  
  This scanner does nothing else, like sets the DROP flag.
  Email is passed to parent scanners and is dropped in sagator.
  '''
  name='drop()'
  def __init__(self,drop_pattern,*scanners):
      self.ONLY_IF=drop_pattern
      self.DROP_LEVEL=9999.9
      match_any.__init__(self,scanners)
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if re.search(self.ONLY_IF,detected,re.IGNORECASE):
        globals.ACTION=S_DROP
      if level>=self.DROP_LEVEL:
        globals.ACTION=S_DROP
      return level,detected,virlist

class deliver(match_any):
  '''
  deliver() - a class to force sending some viruses/spams

  usage: deliver(deliver_pattern,scanners)
  
  where: deliver_pattern is an regular expression, which defines
           virus name to really deliver, other emails will not be delivered.

  This scanner can be used to send some viruses/spams.
  You can define by regular expression, what virus name or spam
  can be sent.
  '''
  name='deliver()'
  def __init__(self,*scanners):
      self.ONLY_IF='.'
      match_any.__init__(self,scanners)
  def onlyif(self,pattern):
      self.ONLY_IF=pattern
      return self
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if re.search(self.ONLY_IF,detected,re.IGNORECASE):
        globals.ACTION=S_FORCE_SEND
      return level,detected,virlist

class deliver_to(match_any):
  '''
  deliver_to() - an interscanner to send emails to admins
  
  Usage: deliver_to(recipients,scanners)
     or: deliver_to(...).onlyif('string')
  
  Where: recipients is an array of recipients
         'string' is an string, which defines regular expression.
           If this expression is found in virus name, mail is
           delivered, otherwise don't.

  This scanner can be used to send viruses/spams to an administrator.
  '''
  name='deliver_to()'
  def __init__(self,recipients,*scanners):
      self.RECIPIENTS=recipients
      self.ONLY_IF='.'
      match_any.__init__(self,scanners)
  def onlyif(self,pattern='.'):
      self.ONLY_IF=pattern
      return self
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if globals.scan_only:
        return level,detected,virlist
      if re.search(self.ONLY_IF,detected,re.IGNORECASE):
        if self.RECIPIENTS:
          try:
            debug.echo(4,"deliver_to(): Trying to deliver to: ",
                         str(self.RECIPIENTS))
            smtpc().sendmail(mail.sender,self.RECIPIENTS,
              mail.header+mail.xheader+mail.data[mail.bodypos:])
          except:
            debug.echo(1,"deliver_to(): An error occured when sending reports")
            debug.traceback(4,"deliver_to():")
            raise ScannerError,"deliver_to()"
      return level,detected,virlist



'''
email parser module, version 0.6.1

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>,
              A'rpi <arpi(at)mplayerhq.hu>

Some parts are based on python's email package written by:
              Ben Gertzfield <che(at)debian.org>
              Barry Warsaw <barry(at)zope.com>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
'''

from avlib import *
import os,re,filetype
import mimetypes
import binascii,uu,base64,binhex,quopri
from StringIO import StringIO

__all__=['parsemail']

def hdr_parse(s,pmode=1):
    params=[]
    qflag=iflag=icnt=cflag=0
    comment=key=value=""
    keymode=1
    lastnonws=0
    wsflag=1
    for c in s:
	if not qflag:
	    # process control chars, by priority:
	    if c=='\\':
		qflag=1
		continue;
	    if c=='"':
		iflag=not iflag
		icnt+=1
		wsflag=0
		continue;
	    if not iflag:
		if wsflag:
		    # skip initial whitespaces, unless they're quoted
		    if c in ('\t',' '):
			continue
		    wsflag=0
		if c=='(':
		    cflag+=1
		    continue
		if c==')':
		    cflag-=1
		    continue
		if cflag==0 and pmode:
		    if c=='=' and keymode:
			keymode=0
			lastnonws=0
			wsflag=1
			continue
		    if c==';':
			params.append((key.strip().lower(),value[:lastnonws],comment.strip()))
			comment=""
			key=""
			value=""
			keymode=1
			icnt=0
			continue
	if cflag>0:
	    comment+=c
	else:
	    if keymode:
		key+=c
	    else:
		value+=c
		if iflag or qflag or not c in ('\t',' '):
		    lastnonws=len(value)
	qflag=0

    if not pmode:
	return (key.strip(),comment.strip())
    params.append((key.strip().lower(),value[:lastnonws],comment.strip()))
    return params

# inspired by Header.py::__unicode__()
ecre = re.compile(r'''
  =\?                   # literal =?
  (?P<charset>[^?]*?)   # non-greedy up to the next ? is the charset
  \?                    # literal ?
  (?P<encoding>[qb])    # either a "q" or a "b", case insensitive
  \?                    # literal ?
  (?P<encoded>.*?)      # non-greedy up to the next ?= is the encoded string
  \?=                   # literal ?=
  ''', re.VERBOSE | re.IGNORECASE)

def decode_header(header):
    """Decode a message header value without converting charset.

    Returns a list of (decoded_string, charset) pairs containing each of the
    decoded parts of the header.  Charset is None for non-encoded parts of the
    header, otherwise a lower-case string containing the name of the character
    set specified in the encoded string.

    An email.Errors.HeaderParseError may be raised when certain decoding error
    occurs (e.g. a base64 decoding exception).
    """
    # If no encoding, just return the header
    header = str(header)
    if not ecre.search(header):
        return [(header, None)]
    decoded = []
    dec = ''
    for line in header.splitlines():
        # This line might not have an encoding in it
        if not ecre.search(line):
            decoded.append((line, None))
            continue
        parts = ecre.split(line)
        while parts:
            unenc = parts.pop(0).strip()
            if unenc:
                # Should we continue a long line?
                if decoded and decoded[-1][1] is None:
                    decoded[-1] = (decoded[-1][0] + SPACE + unenc, None)
                else:
                    decoded.append((unenc, None))
            if parts:
                charset, encoding = [s.lower() for s in parts[0:2]]
                encoded = parts[2]
                dec = None
                if encoding == 'q':
                    dec = quopri.decodestring(encoded)
                elif encoding == 'b':
                    try:
                        dec = base64.decodestring(encoded)
                    except binascii.Error:
                        raise HeaderParseError
                    except ValueError:
                        raise HeaderParseError
                if dec is None:
                    dec = encoded
                if decoded and decoded[-1][1] == charset:
                    decoded[-1] = (decoded[-1][0] + dec, decoded[-1][1])
                else:
                    decoded.append((dec, charset))
            del parts[0:3]
    return decoded

def hdr_decode(h):
    if not h:
	return u""
    uchunks = []
    last = None
    for s,enc in decode_header(h):
        if last not in (None, 'us-ascii'):
            if enc in (None, 'us-ascii'):
                uchunks.append(u' ')
        else:
	    if enc not in (None, 'us-ascii'):
        	uchunks.append(u' ')
	last=enc
	try:
	    uchunks.append(unicode(s,enc or "latin-1"))
	except:
	    uchunks.append(unicode(s,"latin-1"))
    return u''.join(uchunks)

def plain(s):
    return s

class decode_email:
  def debug(self,level,str):
    if debug.debug_level<5:
      self.max_errors-=1
    if self.max_errors==0:
      debug.echo(level,"parsemail(): TOO MANY ERRORS !!!")
    elif self.max_errors>0:
      debug.echo(level,str)
  def scan_part(self,filename=None):
    stream=self.part.getvalue()
    if not stream:
      return
    if filename:
      self.scanner.filename=normalize_filename(filename)
      self.files.append(self.scanner.filename)
    else:
      self.scanner.filename='unknown.exe'
    debug.echo(5,"parsemail(): stream len=",len(stream),
                 ", filename=",self.scanner.filename)
    debug.echo(9,"decoded data part: "+str([stream[0:40]]),"...")
    if self.scanner.is_multistream:
      self.scanner.addstream(stream)
    else:
      self.scanner.prescan()
      l,v,r=self.scanner.scanstream(stream)
      self.scanner.postscan(l,v,r)
      if not self.vir:
        self.vir=v
      self.ret+=r
      self.level+=l
    # delete old part and create a new StringIO
    self.part=StringIO()
  def __init__(self,data,scanner):
    '''
    Decode whole email and run a scanner on decoded parts.
    '''
    self.level,self.vir,self.ret,self.files=0.0,'',[],[]
    self.scanner=scanner
    in_header=1
    bounds={}
    header=[]
    self.lineno=0
    preline=''
    self.part=StringIO()
    file_name=''
    dio=StringIO(data)
    self.max_errors=20
    while 1:
      line=dio.readline()
      if not line: break
      line=line.rstrip("\r\n")
      self.lineno+=1
      while line[0:2]=="> ":
        line=line[2:]
      if in_header:
        # if line is (really) empty => EOH
        if not line:
          self.scan_part(file_name)
          file_type,file_name,coding=None,None,None
          try:
            p_ctype=p_cdisp=p_ctenc=None
            for hs in header:
              fpos=hs.find(':')
              if fpos<=0:
                # not a field, check if it's a boundary!
                b_ok=0
                for b in bounds.keys():
                  if hs.find("--"+b)==0:
                    debug.echo(8,"HDR: found a boundary inside a header!")
                    b_ok=1
                    break
                if not b_ok:
                  self.debug(1,"HDR: Invalid header field: '"+hs+"'")
                continue
              if len(hs)>998:
                self.debug(1,"HDR: Header line too long! (%d chars)" % (len(hs)))
              field=hs[0:fpos].strip().lower()
              fval=hs[fpos+1:]
              # parse content-type: header
              if field=="content-type":
                if p_ctype:
                  self.debug(1,"HDR: MIME error: content-type redefined!")
                p_ctype=hdr_parse(fval)
                file_type=p_ctype[0][0]
                for param in p_ctype:
                  if param[0]=="name":
                    if not file_name:
                      file_name=hdr_decode(param[1])
                  elif param[0]=="boundary":
                    # not sure if we have to strip() it:
                    bounds[param[1].strip()]=1
                file_name=mimetypes.guess_extension(file_type)
                if file_name:
                  file_name="unknown"+file_name
                continue
              # parse content-disposition: header
              if field=="content-disposition":
                if p_cdisp:
                  self.debug(1,"HDR: MIME error: content-disposition redefined!")
                p_cdisp=hdr_parse(fval)
                for param in p_cdisp:
                  if param[0]=="filename":
                    file_name=hdr_decode(param[1])
                continue
              if field=="content-transfer-encoding":
                if p_ctenc:
                  self.debug(1,"HDR: MIME error: transfer-encoding redefined!")
                else:
                  coding=hdr_parse(fval,0)[0]
                continue
          except:
            self.debug(1,"HDR: Exception while MIME header parsing")
            if self.max_errors>0:
              debug.traceback(4,"HDR: ")
          decoder=plain
          if coding:
            coding=coding.strip().lower()
            if coding=="quoted-printable":
              decoder=binascii.a2b_qp
            elif coding=="base64":
              decoder=binascii.a2b_base64
            elif coding in ('x-uuencode', 'uuencode', 'uue', 'x-uue'):
              decoder=binascii.a2b_uu
            elif file_type=="application/mac-binhex40":
              decoder=binascii.a2b_hqx
          checkend=0
          in_header=0
          header=[]
          self.part=StringIO() # create new part
          continue
      else: # not in_header
        # check for boundaries
        for b in bounds.keys():
          if line.find("--"+b)>=0:
            in_header=1
            debug.echo(8,"HDR: boundary found, switching to in_header")
            continue
        # handle forwarded/bounced MIME emails
        lline=line.strip().lower()
        fpos=lline.find(':')
        if fpos>0:
          field=lline[0:fpos]
          if field in ("content-type","content-transfer-encoding","content-disposition","content-id"):
            in_header=1
            debug.echo(7,"HDR: looks like header, switching to in_header")
        if lline[0:40]=="(this file must be converted with binhex":
          decoder=a2b_hqx
          continue
        if lline[0:6]=="begin " or lline[0:13]=="begin-base64 ":
          try:
            begin,mode,name=lline.split(' ',2)
            if int(mode,8)>0 and name:
              file_name=name.strip()
            if begin.lower()=="begin":
              decoder=binascii.a2b_uu
            else:
              decoder=binascii.a2b_base64
            self.scan_part(file_name)
            self.part=StringIO() # create new part
            checkend=1
            continue
          except:
            pass
      if in_header:
        # do the header un-folding:
        if line[0:1] in [' ','\t']:
          if header:
            header[-1]+=line
          else:
            self.debug(1,"syntax error: bad unfolded header line: '"+line+"'")
        else:
          header.append(line)
      else:
        # decode
        try:
          preline+=line
          if preline:
            self.part.write(decoder(preline))
            preline=''
        except:
          if checkend and line.lower()[0:3]=="end":
            self.scan_part(file_name)
            decoder=None
            self.part=StringIO()
            checkend=0
            preline=''
            continue
          if line[-2:]=='=9':
            # correct end
            preline=''
            continue
          self.debug(3,
            "parsemail(): Exception while attachment decoding, file may be truncated! length=%d, line[%d]=...'%s'" \
              % (len(preline),self.lineno,preline[-80:]))
    self.scan_part(file_name)
  def scanstream(self):
    if self.scanner.is_multistream:
      debug.echo(5,"Scanning as multistream")
      self.level,self.vir,self.ret=self.scanner.scanmultistream([])
    del self.scanner
    return self.level,self.vir,self.ret

class parsemail(interscanner):
      '''
      mimeparse interscanner
      
      Usage: parsemail(scanner1(...))
      
      This scanner parses emails and send multiple of streams into
      scanner1(). If an multiscanner is used as another_scanner,
      optimal method is used (scans all files at once).
      '''
      name='ParseMail()'
      def __init__(self,scanner):
          self.name="ParseMail("+scanner.name+")"
          self.scanner=scanner
      def scanstream(self,stream):
          level,vir,ret=decode_email(stream,self.scanner).scanstream()
          return level,vir,ret



'''
Advanced logger for sagator, version 0.4
 
(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
'''
                                                                                
from avlib import *
from match import match_any
from report import report
                                                                                
__all__=['log','log_mysql','log_pgsql','log_sqlite']

class log(report):
  '''
  Advanced logger interscanner.
  
  usage: log(logto,format,scanners...)
  
  where: logto is a string, which defines filename to store these data
           If logto is an integer, it defines log level in standard
           log file.
         format is a string, which defines data format

  This scanner can be used to log some special data.
  You can use these variables in format string:
    $LEVEL		- detected virus level
    $VIRNAME		- name of detected virus, empty if CLEAN
    $STATUS		- if mail is dropped, rejected, ...
    $QNAME		- quarantine file name
    $SCANNER_NAME	- scanner which reported this virus
    $SENDER		- scanned message sender
    $RECIPIENTS		- scanned message recipients
    $SIZE               - email size
    $VERSION		- sagator's version
  
  You also can use log.FORMAT for default format.
  '''
  name='log()'
  # default format
  FORMAT="level='$LEVEL', virname='$VIRNAME', size='$SIZE', sender='$SENDER', recipients='$RECIPIENTS'\n"
  def __init__(self,logto,format,*scanners):
      self.logto=logto
      self.format=format
      match_any.__init__(self,scanners)
  def quote(self,s):
      return quote(s)
  def repl_vars(self,level,detected):
      return self.replace_tmpl(self.format,{
        'LEVEL':	str(level),
        'VIRNAME':	self.quote(detected),
        'STATUS':	globals.action(level),
        'QNAME':	self.quote(globals.QFNAME),
        'SCANNER_NAME':	globals.found_by.name,
        'SENDER':	self.quote(mail.sender.strip()),
        'RECIPIENTS':	self.quote(string.join(mail.recip,',')),
        'SIZE':		str(len(mail.data)),
        'VERSION':	SG_VER_REL
      })
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if type(self.logto)==type(1):
        debug.echo(self.logto,self.repl_vars(level,detected).strip())
      else:
        open(self.logto,'a').write(self.repl_vars(level,detected))
      return level,detected,virlist

class _log_sql(log):
  '''
  SQL interscanner for python DB-API 2.0 compatible DB modules.
  
  This scanner is used as parent for various SQL based log_scanners.
  '''
  FORMAT="INSERT INTO sagator " \
    +"(datetime,level,virname,status,qname,sender,recipients,size)" \
    +" VALUES " \
    +"(now(),'$LEVEL','$VIRNAME','$STATUS','$QNAME','$SENDER','$RECIPIENTS','$SIZE');"
  def __init__(self,sqldb,format,*scanners):
      self.sqldb=sqldb
      self.format=format
      self.dbc=None
      self.connect()
      match_any.__init__(self,scanners)
  def __del__(self):
      try:
        self.dbc.commit()
      except:
        pass
  def quote(self,s):
      return s.replace("'","''").replace('\\','\\\\')
  def connect(self):
      raise 'You need to define connect() for SQL logger!'
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      for TRY in [1,2]:
        try:
          self.connect()
          cur=self.dbc.cursor()
          insert=self.repl_vars(level,detected)
          debug.echo(5,"log_sql(): ",str([insert]))
          cur.execute(insert)
          self.dbc.commit()
        except:
          if TRY==1:
            debug.echo(3,"log_sql(): WARNING: Problem logging into database!")
            debug.traceback(5,"log_sql()")
          else:
            debug.echo(1,"log_sql(): ERROR: Error logging into database!")
            debug.traceback(4,"log_sql()")
          # next connect will be created on next email
          self.dbc=None
      return level,detected,virlist

class log_mysql(_log_sql):
  '''
  MySQL logger interscanner.
  
  usage: log_mysql(host, port, dbname,dbuser,dbpass,format,scanners...)
  
  where: host is a string, which defines MySQL database hostname
           (you can use "127.0.0.1", UNIX socket can't be used in chroot)
         port is an integer, which defines MySQL database TCP port number
           (it is by default 3306)
         dbuser is a string, which defines username which can log to DB
         dbname is a string, which defines database name
         dbpass is a string, which defines user password
         format is a string, which defines data format.
           For detailed format description see log() scanner documentation.
           You also can use log.FORMAT for default format.
  '''
  def __init__(self,host,port,dbname,dbuser,dbpasswd,format,*scanners):
      try:
        import MySQLdb
      except ImportError,es:
        sys.exit("Error importing MySQLdb module for python: "+str(es))
      self.dbhost=host
      self.dbport=port
      self.dbuser=dbuser
      self.dbname=dbname
      if dbpasswd:
        self.dbpasswd=dbpasswd
      else:
        self.dbpasswd=''
      _log_sql.__init__(self,MySQLdb,format,scanners)
  def connect(self):
      if not self.dbc:
        # make new connection to database
        debug.echo(7,"log_mysql: Connecting to DB...")
        self.dbc=self.sqldb.connect(
          host=self.dbhost,
          user=self.dbuser,
          db=self.dbname,
          port=self.dbport,
          passwd=self.dbpasswd,
          connect_timeout=5
        )

class log_pgsql(_log_sql):
  '''
  PosgreSQL logger interscanner.
  
  usage: log_pgsql(host, port, dbname,dbuser,dbpass,format,scanners...)
  
  where: host is a string, which defines pgsql database hostname
           (you can use "localhost", UNIX socket can't be used in chroot)
         port is an integer, which defines pgsql database TCP port number
           (it is by default 5432)
         dbuser is a string, which defines username which can log to DB
         dbname is a string, which defines database name
         dbpass is a string, which defines user password
         format is a string, which defines data format.
           For detailed format description see log() scanner documentation.
           You also can use log.FORMAT for default format.
  '''
  FORMAT="INSERT INTO sagator VALUES ( now(), '$LEVEL', '$VIRNAME', '$STATUS', '$QNAME', '$SENDER', '$RECIPIENTS', '$SIZE' )"
  def __init__(self,host,port,dbname,dbuser,dbpasswd,format,*scanners):
      try:
        import pg
      except ImportError,es:
        sys.exit("Error importing PostgreSQL module for python: "+str(es))
      self.dbhost=host
      self.dbport=port
      self.dbuser=dbuser
      self.dbname=dbname
      self.dbpasswd=dbpasswd
      _log_sql.__init__(self,pg,format,scanners)
  def connect(self):
      if not self.dbc:
        # make new connection to database
        self.dbc=self.sqldb.connect(
          dbname=self.dbname,
          host=self.dbhost,
          port=self.dbport,
          user=self.dbuser,
          passwd=self.dbpasswd
        )

class log_sqlite(_log_sql):
  '''
  SQLite logger interscanner.
  
  usage: log_sqlite(dbname,format,scanners...)
  
  where: dbname is a string, which defines, which database to use
         format is a string, which defines data format.
           For detailed format description see log() scanner documentation.
           You also can use log.FORMAT for default format.
  '''
  def __init__(self,dbname,format,*scanners):
      try:
        import sqlite
      except ImportError,es:
        sys.exit("Error importing SQLite module for python: "+str(es))
      self.dbname=dbname
      _log_sql.__init__(self,sqlite,re.sub("now\(\)","0",format),scanners)
  def connect(self):
      if not self.dbc:
        self.dbc=self.sqldb.connect(safe.fn(self.dbname))



'''
Report interscanners for sagator, version 0.6.2

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import re,string,socket
from match import match_any

__all__=['report','report_recipients']

#####################################################################
### REPORT class

# MESSAGE REPORT TEMPLATE
REPORT_MSG_TEMPLATE='''\
From: $FROM1
To: $TO1
Subject: ALERT, $VIRNAME1 FOUND!

                             VIRUS FOUND!
---------------------------------------------------------------------
STATUS: $STATUS1
QUARANTINED AS: $QNAME1
---------------------------------------------------------------------
SMTP COMMUNICATION:
$SMTP_COMM1
---------------------------------------------------------------------
HEADER OF MESSAGE:
$MSG_HEADER1
---------------------------------------------------------------------
SCANNER: $SCANNER_NAME1
$SCANNER_OUTPUT1
'''

class report(match_any):
  '''
  report() interscanner - report any message to admin, user or anybody
  
  usage: report(recipients, msg, scanners)
  or:    report(...).sender('<email@of.sender.dom>')
  or:    report(...).ifscan(cond_scanner)

  where: recipients is an array of recipients,
           like: ['<r1@dom>','<r2@dom>',...]
         msg is an message template, you can use report.MSG_TMPL
         cond_scanner is a scanner, which defines, when report may be sent

  This module can be used to report an message to administrator
  or to any user.

  You can use following variables in report messages:
    $FROM		report sender
    $VIRNAME		detected virus name
    $SMTP_COMM		SMTP communication
    $STATUS		status of this message (REJECTED, DROPPED, ...)
    $QNAME		filename in quarantine
    $SCANNER_OUTPUT	returned scanner output
    $SCANNER_NAME	name of the scanner
    $MSG_HEADER		virus/spam header
    $MSG_BODY		virus/spam body (without header)
    $RANDOM		10 random characters (useable to generate a boundary)
    $SENDER		virus/spam sender
    $RECIPIETNS		virus/spam recipients joined by comma (',')
    $TO			report recipeints
    $VERSION            sagator's version
  '''
  # Templates
  MSG_TMPL=REPORT_MSG_TEMPLATE
  # scanner
  name='report()'
  def __init__(self,recipients,msg,*scanners):
      self.MSG=msg
      self.RECIPIENTS=recipients
      self.SENDER='<vscan@'+socket.gethostname()+'>'
      match_any.__init__(self,scanners)
  def sender(self,email):
      self.SENDER=email
      return self
  def replace_tmpl(self,msg,vars):
      idx={}
      for v in vars.keys():
        try:
          idx[msg.index('$'+v)]=v
        except ValueError:
          pass
        for c in '0123456789':
          try:
            idx[msg.index('$'+v+c)]=v+c
          except ValueError:
            pass
      keys=idx.keys() # replace it from last to first
      keys.sort()
      keys.reverse()
      for k in keys:
        idxk=idx[k]
        if idxk[-1] in '0123456789':
          idxk=idxk[:-1]
        msg=msg[:k]+vars[idxk]+msg[k+len(idx[k])+1:]
      return msg
  def ifscan(self,scanner):
      self.IFSCAN=scanner
      return self
  def ifscan_main(self,stream):
      try:
        level,detected,virlist=self.IFSCAN.scanstream(stream)
      except (NameError,AttributeError):
        return True
      if is_infected(level,detected):
        debug.echo(6,'report(): condition: True (%f,%s)' % (level,detected))
        return True
      else:
        debug.echo(6,'report(): condition: False (%f,%s)' % (level,detected))
        return False
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if globals.scan_only:
        return level,detected,virlist
      if not self.ifscan_main(stream):
        return level,detected,virlist
      if not is_infected(level,detected):
        return level,detected,virlist
      if self.RECIPIENTS:
        repl_vars={
          'FROM':		self.SENDER,
          'VIRNAME':		detected,
          'SMTP_COMM':		mail.addr+mail.comm.rstrip(),
          'STATUS':		globals.action(level),
          'QNAME':		globals.QFNAME,
          'SCANNER_OUTPUT':	string.join(virlist,""),
          'SCANNER_NAME':	globals.found_by.name,
          'MSG_HEADER':		mail.header.rstrip(),
          'MSG_BODY':		mail.data[mail.bodypos:],
          'RANDOM':		randomchars(10),
          'SENDER':             mail.sender,
          'RECIPIENTS':         string.join(mail.recip,','),
          'TO':			string.join(self.RECIPIENTS,',\r\n\t'),
          'VERSION':		SG_VER_REL
        }
        msg=self.replace_tmpl(self.MSG,repl_vars)
        try:
          smtpc().sendmail(self.SENDER,self.RECIPIENTS,msg)
        except SmtpcError,(ses,es):
          debug.echo(1,"report(): AdminReport: ",ses)
      return level,detected,virlist

class report_recipients(report):
  '''
  report_recipients() interscanner - report any message to mail recipients
 
  usage: report_recipients(msg, scanners)
  or:    report_recipients(...).sender('<email@of.sender.dom>')
  or:    report(...).ifscan(cond_scanner)

  where: msg is an message template, you can use report.MSG_TMPL
          scanners are scanners which are used
         cond_scanner is a scanner, which defines, when report may be sent

  This module can be used to report an message to mail recipients.
  For message description see report() scanner.
  '''
  # Templates
  MSG_TMPL=REPORT_MSG_TEMPLATE
  # scanner
  name='report()'
  def __init__(self,msg,*scanners):
      self.MSG=msg
      self.SENDER='<vscan@'+socket.gethostname()+'>'
      match_any.__init__(self,scanners)
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      if globals.scan_only:
        return level,detected,virlist
      if not self.ifscan_main(stream):
        return level,detected,virlist
      if not is_infected(level,detected):
        return level,detected,virlist
      repl_vars={
          'FROM':		self.SENDER,
          'VIRNAME':		detected,
          'SMTP_COMM':		mail.addr+mail.comm.rstrip(),
          'STATUS':		globals.action(level),
          'QNAME':		globals.QFNAME,
          'SCANNER_OUTPUT':	string.join(virlist,""),
          'SCANNER_NAME':	globals.found_by.name,
          'MSG_HEADER':		mail.header.rstrip(),
          'MSG_BODY':		mail.data[mail.bodypos:],
          'RANDOM':		randomchars(10),
          'SENDER':             mail.sender,
          'RECIPIENTS':         string.join(mail.recip,','),
          'TO':			string.join(mail.recip,',\r\n\t'),
          'VERSION':		SG_VER_REL
      }
      msg=self.replace_tmpl(self.MSG,repl_vars)
      try:
        smtpc().sendmail(self.SENDER,mail.recip,msg)
      except SmtpcError,(ses,es):
        debug.echo(1,"report(): UserReportRecipients: ",ses)
      return level,detected,virlist



'''
Sagators statistics collector and distributor, version 0.3.1

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

Examples:
  python2 sagator.py /
  python2 sagator.py mrtg Total-count Filtered-count

'''

import socket,re,string,os,sys
import urllib,time,signal,struct
import interscan.match
from avlib import *

__all__ = ['collector','update','query','status']

COLLECTOR_SERVER=()
MYCOLLECTOR=None

def uptime(starttime):
    up=time.mktime(time.localtime())-starttime
    mins=int(up/60)%60
    hours=int(up/3600)%24
    days=int(up/86400)
    if days>0:
      return "%d days, %02d:%02d" % (days,hours,mins)
    else:
      return "%02d:%02d" % (hours,mins)

class collector:
  pid=0
  data={ 'Total-count':0.0, 'Total-bytes':0.0,
         'Filtered-count':0.0, 'Filtered-bytes':0.0,
         'Total-time':0.0 }
  def __init__(self,bindaddr):
      global MYCOLLECTOR
      MYCOLLECTOR=self
      # startup initialization
      self.bindaddr=bindaddr
      # create a server
      self.s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      self.s.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
      self.s.bind(bindaddr)
      self.s.listen(10)
  def main(self):
      global COLLECTOR_SERVER
      COLLECTOR_SERVER=self.bindaddr
      self.lastsave=time.mktime(time.localtime())
      self.pipe_r,self.pipe_w=os.pipe() # pipe for updates
      self.pid=os.fork()
      if self.pid==0: # main process
        # chroot
        from aglib import dochroot,sighup
        dochroot()
        signal.signal(signal.SIGHUP,sighup)
        signal.signal(signal.SIGUSR1,self.sigusr1)
        signal.signal(signal.SIGTERM,self.sigterm)
        globals.pidf.close()
        os.close(self.pipe_w)
        debug.dup()
        self.data['Start-at']=self.lastsave
        self.loadstats()
        self.data['Restart-at']=self.lastsave
        self.data['Hostname']=socket.gethostname()
        debug.echo(1,"collector(): service started, waiting for connections...")
        self.mypid=os.getpid()
        while 1:
          try:
            self.conn,addr=self.s.accept()
            socket_settimeout(self.conn,8)
            try:
              f=self.conn.makefile("rw",0)
              req=f.readline()
              debug.echo(5,"collector(): QUERY: ",req.strip())
              proto=0.0
              reg1=re.search("^GET /([^ ]*) HTTP/(1.[01])",req)
              if reg1:
                dirs=reg1.group(1).split("/")
                proto=float(reg1.group(2))
              else:
                reg1=re.search("^GET /([^ ]*)$",req.rstrip())
                if reg1:
                  proto=1.0
                  dirs=reg1.group(1).split("/")
              if proto>=1.0:
                args=self.parseargs(dirs[1:])
                lines={}
                while proto>=1.1:
                  line=f.readline()
                  if (line=="\r\n") or (line=="\n"):
                    break
                  a=line.split(": ")
                  if len(a)>1:
                    lines[a[0]]=a[1].strip()
                  else:
                    lines[a[0]]=a[0].strip()
                self.conn.sendall("HTTP/1.1 200 OK\r\n")
                self.conn.sendall("Content-Type: text/plain; charset=iso-8859-1\r\n")
                self.conn.sendall("\r\n");
                self.data["Uptime"]=uptime(self.data['Start-at'])
                self.data["Restart-age"]=uptime(self.data['Restart-at'])
                sstr=''
                if dirs[0]=="":
                  for key,value in self.data.items():
                    sstr+=key+":\t"+str(value)+"\r\n"
                  self.conn.sendall(sstr)
                elif dirs[0]=="mrtg":
                  for key in dirs[1:3]+['Uptime','Hostname']:
                    if key in self.data.keys():
                      try:
                        o="%-20.0f" % self.data[key]
                        sstr+=o.strip()+"\r\n"
                      except TypeError:
                        sstr+=str(self.data[key])+"\r\n"
                    else: # return 0 for unknown key
                      sstr+="0\r\n"
                  self.conn.sendall(sstr)
                else:
                  self.message("404 Not Found")
              else:
                self.message("501 Method Not Implemented")
                self.conn.sendall(req)
            except socket.error,(ec,es):
              if not ec in [32,104,107]:
                debug.echo(3,"collector: socket.error: ",ec,es)
                debug.traceback(4,"collector: ")
            except:
              debug.traceback(3,"collector: ")
            try:
              self.conn.shutdown(2)
              self.conn.close()
            except socket.error:
              pass 
          except:
            try:
              self.conn.shutdown(2)
              self.conn.close()
            except:
              pass
        # while 1 ... end
      else:
        self.s.close()
        os.close(self.pipe_r)
      return self.pid
  def update(self,dirs):
      d=string.join(dirs,"/")+"\n"+' '*1024
      signal.signal(signal.SIGALRM,self.sigalrm)
      try:
        signal.alarm(2)
        os.write(self.pipe_w,d[:1024])
        signal.alarm(0)
      except OSError,(ec,es):
        signal.alarm(0)
        if ec==4: # Interrupted system call
          debug.echo(3,"collector(): writing to pipe: ",es)
      except:
        debug.echo(1,"collector(): unknown error on update")
        debug.traceback(3,"collector()")
      signal.signal(signal.SIGALRM,signal.SIG_DFL)
      os.kill(self.pid,signal.SIGUSR1)
  def sigalrm(self,sn,stack):
      # ignore this signal, only do an "interrupted system call"
      debug.echo(1,"collector(): WARNING: Interrupting write to pipe! (pipe full?)")
  def sigusr1(self,sn,stack):
      signal.signal(signal.SIGALRM,self.sigalrm)
      try:
        signal.alarm(2)
        str=os.read(self.pipe_r,1024)
        signal.alarm(0)
      except OSError,(ec,es):
        signal.alarm(0)
        if ec==4: # Interrupted system call
          debug.echo(3,"collector(): ",es)
          return
      signal.signal(signal.SIGALRM,signal.SIG_DFL)
      args=self.parseargs(str.split("/"))
      self.data["Childrens"]=0.0
      for k,v in args.items():
        try:
          self.data[k]+=float(v)
        except KeyError:
          self.data[k]=float(v)
        except ValueError:
          self.data[k]=v
        except TypeError:
          debug.echo(0,"collector(): TypeError: ",k,v,type(v))
      # save stats evry 5 minutes
      ctime=time.mktime(time.localtime())
      if (ctime-self.lastsave)>300:
        self.lastsave=ctime
        self.savestats()
  def sigterm(self,sn,stack):
      try:
        os.close(self.pipe_r)
        debug.echo(4,"collector(): Saving stats...")
        self.savestats()
      except:
        pass
      debug.echo(1,"collector(): Exiting - SIGTERM ...")
      os._exit(0)
  def loadstats(self):
      try:
        f=open('/tmp/status','r')
        for line in f.readlines():
          a=line.split(":\t")
          try:
            self.data[a[0]]=float(a[1])
          except ValueError:
            self.data[a[0]]=a[1]
        f.close()
      except IOError,(ec,es):
        debug.echo(1,"collector(): loadstat error: ",es)
      except:
        debug.echo(1,"collector(): Can't load status file.")
        debug.traceback(1,"collector()")
  def savestats(self):
      f=open('/tmp/status','w')
      for k,v in self.data.items():
        f.write(k+":\t"+str(v)+"\n")
      f.close()
      debug.echo(4,"collector(): Status saved.")
  def message(self,str):
      self.conn.sendall("HTTP/1.1 "+str+"\r\n\r\n")
  def parseargs(self,dirs):
      args={}
      for arg in dirs:
        a=arg.split("=")
        k=urllib.unquote_plus(a[0])
        if re.search("^[A-Za-z0-9-]*$",k):
          if len(a)>1:
            args[k]=urllib.unquote_plus(a[1])
          else:
            args[k]='1'
        else:
          debug.echo(1,"collector(): Unknown key: ",k)
      return args

def query(qt,args):
    if COLLECTOR_SERVER:
      s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(s,10)
      s.connect(COLLECTOR_SERVER)
      urlargv=''
      for arg in args:
        urlargv+="/"+urllib.quote_plus(arg,"/=")
      s.sendall("GET /"+qt+urlargv+" HTTP/1.1\r\n\r\n")
      f=s.makefile("r",0)
      lines=[]
      for _ in range(512):
        line=f.readline()
        if (line=="\r\n") or (line=="\n") or (line==""):
          break;
        lines.append(line)
      lines=[]
      str=''
      str+=s.recv(9999).rstrip()
      s.shutdown(2)
      return str
    else:
      return None

def update(bs=-1,t=0,infected=0,tempfail=0):
    '''update statistics by 'bs' bytes, 't' time,
       if infected, then update also filtered bytes,files
    '''
    args=['Total-count','Total-bytes='+str(bs),'Total-time='+str(t),
          'Temp-fail='+str(tempfail),'Childrens=0']
    if infected:
      args.extend(['Filtered-count','Filtered-bytes='+str(bs)])
    try:
      MYCOLLECTOR.update(args+status.collect)
      status.collect=[]
    except:
      pass

class status(interscan.match.match_any):
  '''
  status() interscanner
  
  Usage: status("String",scanner1(),scanner2(),...)
  
  Where: "String" is a string, which defines a prefix for status update
         scanner1(),... are sagator's scanners
  
  This interscanner can be used to collect some other statistics.
  Preffered usage:
    SCANNERS=[
      status("Virus",virus_scanner1(),...),
      status("Spam",spam_scanner1(),...)
    ]
  This collects Virus-count,Virus-bytes,Spam-count,Spam-bytes in collector.
  '''
  name='' # making this scanner invisible
  collect=[]
  def __init__(self,string,*scanners):
      self.status_string=string
      interscan.match.match_any.__init__(self,scanners)
  def scanstream(self,stream):
      level,vir,ret=interscan.match.match_any.scanstream(self,stream)
      if vir:
        status.collect.extend([self.status_string+"-count",
          self.status_string+"-bytes="+str(len(stream))])
      return level,vir,ret

if __name__ == '__main__':
  if len(sys.argv)>1:
    # get host and port from collector service
    from config import SRV
    for service in SRV:
      if service.name=='collector()':
        COLLECTOR_SERVER=(service.HOST,service.PORT)
    print query(sys.argv[1],sys.argv[2:])



'''
qsf module, version 0.1

(c) 2003 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import re

__all__=['qsf']

class qsf(ascanner):
  '''
  Quick Spam Filter scanner
  
  Usage: qsf(['/usr/bin/qsf','-r'],SPAMSTRING,MAX_FILE_SIZE)
  
  Where: in [] is a path and argumets of qsf binary
         SPAMSTRING is a string, which define a report string,
           which is checked. If it is found, mail is assigned as 'SPAM'
         MAX_FILE_SIZE is a number, which defines maximum file size,
           which can be tested. If a mail (with header) is larger,
           it is not checked (by default 500000).

  This scanner scans for spams. If a spam is returned, a string
  SPAM is returned as virus name. It uses Quick Spam Filter.
  '''
  is_spamscan=1
  name='qsf()'
  def __init__(self,arg,spamstring='',max_file_size=500000):
      self.arg=arg
      self.SPAMSTRING=spamstring
      self.MAX_FILE_SIZE=max_file_size
  def scanstream(self,stream):
      exitstatus=''
      if len(stream)>self.MAX_FILE_SIZE:
        return 0.0,'',['File too long to test it. Size: '+str(len(stream))]
      pf=popen(self.arg)
      pf.write(stream)
      pf.tocmd.close()
      detected=''
      ret=[]
      level=0.0
      for line in pf.readlines():
        if line.rstrip=="":
          break
        reg1=re.search("^X-Spam: YES$",line,re.IGNORECASE)
        if reg1:
          detected='SPAM'
          mail.addheader(line)
          ret.append(line)
        else:
          reg2=re.search("^X-Spam-Rating: (.*)$",line)
          if reg2:
            if detected:
              detected+=":"+reg2.group(1)
            level=float(reg2.group(1))
            mail.addheader(line)
            ret.append(line)
      pf.wait()
      debug.echo(4,"qsf: ",ret)
      return level,detected,ret



'''
avg module, version 0.5.2

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import socket,re,string

__all__=['avgd']

class avgd(ascanner):
  '''
  avgd realscanner
  
  Usage: avgd([[[params],connto],chroot])

  Where: params is a string which defines command line parameters,
           like -arc, -heur, -rt, ... see man page for avgscan
           By default "-arc -heur -rt" is used now.
         connto is a tuple of host,port of avgd, by default ('localhost',55555)
         chroot is an string in which sagator is running, if avgd
           is not running in this chroot

  This scanners is a realscanner, which can be used to scan
  for viruses. It uses AVGd (a deamonized linux version of AVG7).

  To start AVGd in chroot, move all neccessary files into your chroot,
  and change following line in init.d/avgd from:
    daemon $AVG_HOME/avgscan -d
  to:
    daemon chroot /var/spool/vscan $AVG_HOME/avgscan -d
  or similiar.
  '''
  name='avgd()'
  def __init__(self,params='-arc -rt -heur',connto=('localhost',55555),chroot=''):
      self.params=params
      self.connto=connto
      self.chroot=chroot
  def scanfile(self,files,dir=''):
      level,detected,virlist=0.0,'',[]
      s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(s,120)
      try:
        s.connect(self.connto)
      except socket.error,(ec,es):
        debug.echo(2,"ERROR: avgd(): ",es)
        raise
      f=s.makefile('rw',512)
      while not re.search("^220 ",f.readline()):
        pass
      s.sendall("PARM "+self.params+"\n")
      l=f.readline()
      if not re.search("^200 ",l):
        raise ScannerError,'Protocol error: '+l
      for fn in files:
        s.sendall("SCAN "+self.chroot+fn+'\n')
        l=f.readline()
        debug.echo(4,self.name+": ",l.rstrip())
        if re.search("^200 ",l):
          pass
        elif re.search("^403 ",l):
          virlist.append(fn+': '+l+'\n')
          reg1=re.search("^403 File .* infected: +(Virus identified|Could be infected) +(.*)$",l)
          if reg1:
            if reg1.group(1)[0]=='V':
              level+=1
            else:
              level+=0.9
            if detected=='':
              detected=reg1.group(2).strip()
        else:
          raise ScannerError,l
      s.sendall("QUIT\n")
      l=f.readline()
      if not re.search("^221",l):
        debug.echo(1,"avgd(): AVG returned wrong exit code for QUIT: ",l)
      f.close()
      s.shutdown(2)
      s.close()
      return level,detected,virlist



# InterScanners

from file import *
from match import *
from actions import *
from report import *
from header import *
from decompress import *
from parsemail import *
from mimeparse import *
from logger import *
from store import *



'''
Antivir Gateway Library for Sagator

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

This program is a email antivir gate for an SMTP server.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

import os,sys,string,re,time
import pwd,grp,socket,rfc822

from avlib import *
import stats

TEST_MSG='''\
From: Test User <test@tester.salstar.sk>\r
To: Test User2 <test2@tester2.salstar.sk>\r
Date: Fri, 16 Apr 2004 15:38:51 +0200 (CEST)\r
Subject: testing mail\r
\r
simple mail\r
'''

#########################################################################
### service class template

class service:
  '''Sagator service template'''
  def daemonize(self):
      # redirect stout and strerr
      debug.dup()
      # fork
      pid=os.fork()
      if pid<=0:
        globals.pidf.close()
      return pid

#########################################################################
### Some helpful functions

# chroot, setgid and setuid
def dochroot():
    '''
    Do chroot() and setuid(),setgid()
    '''
    def func_err(fn,f):
        try:
          eval(f)
          return 0
        except OSError,(ec,es):
          debug.echo(0,"WARNING! Can't do %10s: %s" % (fn,es))
        except AttributeError:
          debug.echo(0,"WARNING! Can't do "+f+": Your python don't have this function!")
        return 1
    uid,gid=0,0
    if globals.USER:
      try:
        uid=pwd.getpwnam(globals.USER)[2]
      except KeyError,es:
        debug.echo(0,"ERROR, getpwnam: ",[[globals.USER]],es)
        sys.exit(100)
    if globals.GROUP:
      try:
        gid=grp.getgrnam(globals.GROUP)[2]
      except KeyError,es:
        debug.echo(0,"ERROR, getgrnam: ",[[globals.GROUP]],es)
        sys.exit(100)
    if globals.CHROOT:
      if func_err('chroot',"os.chroot('"+globals.CHROOT+"')"):
        safe.ROOT_PATH=globals.CHROOT
    if gid:
      func_err('setgroups',"os.setgroups(["+str(gid)+"])")
      func_err('setgid',"os.setgid("+str(gid)+")")
    if uid:
      func_err('setuid',"os.setuid("+str(uid)+")")
    if (os.getuid()==0) or (os.getgid()==0
        or (os.geteuid()==0) or (os.getegid()==0)):
      debug.echo(0,"ERROR! I can't setuid and i am running as root!")
      sys.exit(10)

def test_scanners(scanners):
    for scnr in scanners:
      scnr.reinit()
    if os.fork()==0:
      dochroot()
      globals.scan_only=True
      scanarr,scannames=[],'None'
      mail.data=TEST_MSG
      for scnr in scanners:
        if scnr.name:
          debug.echo(1, "Testing "+scnr.name+"...")
        level,detected,virlist,scan_report,err=do_scan(scnr)
        if scan_report:
          debug.echo(0, "Scanner ",scnr.name," test failed! Disable it manually!")
          debug.echo(0, "  ",scan_report)
        scanarr=scanarr+[scnr.name]
      mail.data=''
      sys.exit(0)
    os.wait()

def do_scan(scnr,fn=''):
    '''
    Scan email (stored in mail class) for viruses.
    Use scnr scanner.
    '''
    level,detected,virlist=0.0,'',[]
    try:
      if fn:
        scnr.filename=fn
      level,detected,virlist=scnr.scanstream(mail.data)
      scnr.destroy()
      return level,detected,virlist,'',None
    except socket.error,(ec,es):
      debug.traceback(4,"do_scan: ")
      return level,detected,virlist,\
        scnr.name+": SocketError: "+es,'Socket error: '+es
    except (IOError,OSError):
      debug.traceback(4,"do_scan: ")
      return level,detected,virlist,\
        scnr.name+": Error: "+debug.traceback_value_str(),'Error: '+debug.traceback_value_str()
    except (ScannerError,ValueError),es:
      debug.traceback(4,"do_scan: ")
      return level,detected,virlist,\
        scnr.name+": "+str(es),'Error: '+str(es)
    except:
      debug.traceback(4,"do_scan: ")
      return level,detected,virlist,scnr.name+": unknown error",'Unknown error'

def checkvir(scanners):
    '''
    Check email (stored in mail class) for viruses.
    '''
    try:
      level,detected,virlist=0.0,'',[]
      scanarr,scannames=[],'None'
      for scnr in scanners:
        l,detected,virlist,scan_reply,err=do_scan(scnr)
        level+=l
        if scan_reply:
          debug.echo(0, "CheckVir: ERROR: ",scan_reply)
          return S_TEMPFAIL,scan_reply
        else:
          scanarr=scanarr+[scnr.name]
          scannames=string.join(scanarr,' ')
        if is_infected(level,detected):
          debug.echo(3,"Status: ",[detected,virlist,level])
          break
      mail.addvirusinfo(detected,scannames)
      if not is_infected(level,detected):
        return S_OK,'CLEAN' # mail delivered OK
      else: # VIRUS FOUND
        rstat=globals.ACTION
        debug.echo(5,"action: ",rstat)
        return rstat,detected
    except socket.error,(ec,es):
      debug.echo(0,"checkvir: ERROR: socket.error: ",[[ec]],es)
      debug.traceback(4,"checkvir: ")
      return S_TEMPFAIL,es
    except IOError,(ec,es):
      debug.echo(0,"checkvir: ERROR: IO: ",es)
      debug.traceback(4,"checkvir: ")
      return S_TEMPFAIL,'IOError'
    except:
      debug.traceback(3,"checkvir: ")
      return S_TEMPFAIL,'AnyError, see sagator logs'

#########################################################################
### Signal handlers

def sigchld(sn,stack):
    ''' signal: Child process terminated '''
    # try to remove older zombies (sigchld lost?)
    try:
      while 1:
        pid,s=os.waitpid(0,os.WNOHANG)
        if pid==0:
          break
        debug.echo(1,"Child exited ... pid,status: ",(pid,s))
    except OSError: # no child processes?
      pass

def sighup(sn,stack):
    ''' signal: HUP signal '''
    debug.echo(3,"SIGHUP: Reopening log ...")
    debug.reopen()

def sigusr2(sn,stack):
    ''' signal: USR2 signal '''
    debug.echo(1,"SIGUSR2: Tracelog request received ...")
    debug.debug_level=debug.trace_level
    debug.stack(1,"SIGUSR2: ")

def sigterm(sn,stack=None):
    ''' signal: Terminate process '''
    if type(sn)==type(''):
      debug.echo(1,"SERVER: Exiting - %s ..." % sn)
    else:
      debug.echo(1,"SERVER: Exiting - SIGTERM ...")
    try:
      os.unlink(globals.pid_file)
    except OSError:
      pass
    try:
      if globals.stat_coll.pid>0:
        os.kill(globals.stat_coll.pid,signal.SIGTERM)
    except:
      pass
    sigchld(None,None) # unzombie all childs
    sys.exit(0)



'''
Store/Restore interscanners for sagator, version 0.1

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
'''
                                                                                
from avlib import *
from match import match_any
                                                                                
__all__=['store','restore']

class store(match_any):
  '''
  Store scanner return value.
  
  Usage: store('VAR',scanners...)
  
  Where: 'VAR' is a string, which defines variable name to store return value

  Example: store('var1', scanner1())
  
  This scanner return's return value for a scanner included into
  and stores it's value into a variable.
  '''
  name='store()'
  VARA={}
  def __init__(self,var,*scanners):
      self.var=var
      match_any.__init__(self,scanners)
  def scanstream(self,stream):
      level,detected,virlist=match_any.scanstream(self,stream)
      store.VARA[self.var]=level,detected,virlist
      return level,detected,virlist

class restore(match_any):
  '''
  Restore scanner return value.
  
  Usage: restore('VAR')
  
  Where: 'VAR' is a string, which defines variable name to store return value

  Example: store('var1')
  
  This scanner return's return value for previously stored variable.
  '''
  name='restore()'
  def __init__(self,var):
      self.var=var
  def scanstream(self,stream):
      return store.VARA[self.var]



#!/usr/bin/python
'''
ICAP module, version 0.2
 
(c) 2003 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
'''

from avlib import *
import socket,re

__all__=['savse']

class icap:
  def __init__(self,host,port,path='AVSCAN?action=SCAN'):
      self.s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(self.s,120)
      self.s.connect((host,port))
      self.f=self.s.makefile('rw',0)
      self.server=host+":"+str(port)
      self.path=path
  def check(self,stream,filename):
      level,virname,response=0.0,'',[]
      # send ICAP request
      self.s.sendall("RESPMOD icap://"+self.server+"/"+self.path+" ICAP/1.0\r\n")
      self.s.sendall("Host: "+self.server+"\r\n")
      self.s.sendall("Allow: 204\r\n")
      req="GET http://sagator.salstar.sk/"+filename+" HTTP/1.1\r\n\r\n"
      hdr="HTTP/1.1 200 OK\r\nTransfer-Encoding: chunked\r\n\r\n"
      self.s.sendall("Encapsulated: req-hdr=0, res-hdr=%d, res-body=%d\r\n\r\n"
        % (len(req),len(req+hdr)))
      debug.echo(7,'icap(): ',req,hdr)
      self.s.sendall(req)
      self.s.sendall(hdr)
      # send data stream
      #i,l,s=0,len(stream),0xFF
      #while True:
      #  if (i+s)>=l:
      #    self.s.sendall("%X\r\n" % (len(stream[i:])))
      #    self.s.sendall(stream[i:])
      #    self.s.sendall("\r\n\r\n0\r\n\r\n")
      #    break
      #  else:
      #    self.s.sendall("%X\r\n" % s)
      #    self.s.sendall(stream[i:i+s])
      #    self.s.sendall("\r\n\r\n")
      #  i+=s
      self.s.sendall("%X\r\n" % (len(stream)))
      self.s.sendall(stream)
      self.s.sendall("\r\n\r\n0\r\n\r\n")
      # read protocol status
      ics=self.f.readline()
      debug.echo(7,"savse():",ics)
      try:
        icc=re.search("^ICAP/[0-9.]* ([0-9]{3}) ",ics).group(1)
      except:
        icc='000'
      if icc=='204':
        pass # it's clean
      elif (icc[:2]=='20'):
        # returned OK
        while 1:
          line=self.f.readline()
          if not line:
            break
          debug.echo(7,"savse():",line)
          reg=re.search("^X-Infection-Found: Type=([0-9]*); Resolution=([0-9]*); Threat=([^;]*);",line)
          if reg:
            if (reg.group(1)=='0'):
              virname+=reg.group(3)
              level+=1.0
              debug.echo(5,"savse(): found:",virname)
          else:
            reg1=re.search("^X-Violations-Found:",line)
            if reg1:
              while 1:
                line=self.f.readline()
                if not line:
                  break
                if (line[0]==' ') or (line[0]=='\t'):
                  response.append(line)
                else:
                  break
      else:
        raise 'e',ics
      return level,virname,response
  def close(self):
      self.s.shutdown(2)

class savse(ascanner):
  '''
  savse() realscanner - symantec antivirus scan engine scanner
  
  Usage: savse('host',port)
  
  Example: parsemail(savse('localhost',1344))
  
  Comments:
    You need to configure ICAP protocol on port 1344 (or another)
    to use this scanner properly.
    These options are suggested in symcscan.cfg:
      BindAddress=127.0.0.1
      Port=1344
      Protocol=ICAP
      ExtensionPolicy=0
      ICAPActionPolicy=SCAN
      ICAPResponse=1
  '''
  name='savse()'
  def __init__(self,host,port=1344):
      self.host=host
      self.port=port
  def scanstream(self,stream):
      c=icap(self.host,self.port)
      level,r,response=c.check(stream,self.filename)
      c.close()
      return level,r,response



'''
avfilter.py - SMTP daemon service for sagator.

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from aglib import *
import sys

__all__=['avfilter']

class avfilter(service):
  '''
  AV filter service for sagator.
  
  This service can be used to filter an email through sagator.
  Some headers will be added to filtered email.
  
  Usage: avfilter(scanners)
  
  Where: scanners is an array of scanners (see README.scanners for more info)
  
  Example: avfilter(SCANNERS)
  
  You can use this service for example with procmail.
  Example .procmailrc:
    # filter through sagator
    :0fw
    |sagator --logfile /var/log/sagator.log
    
    # move identified emails to quarantine
    :0
    * ^X-Sagator-Status: .
    $HOME/mail/quarantine
  '''
  name='avfilter()'
  def __init__(self,scanners):
      self.SCANNERS=scanners
  def main(self):
      for scnr in self.SCANNERS:
        scnr.reinit()
      globals.pid_file=None
      try:
        # read stdin
        while 1:
          line=sys.stdin.readline()
          if not line: break
          mail.df.write(line)
        mail.close()
        stdout=os.dup(1)
        debug.dup()
        status,virname=checkvir(self.SCANNERS)
        os.write(stdout, mail.xheader)
        os.write(stdout, mail.header)
        os.write(stdout, mail.data[mail.bodypos:])
        sys.exit(0)
      except (KeyboardInterrupt,SystemExit):
        raise
      except:
        debug.echo(1,"avfilter(): Unknow error!")
        debug.traceback(0,"avfilter(): ")
        sys.exit(1)



'''
Antivir library for Sagator

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj(at)salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

import os,sys,re,string,time,random
import traceback,signal,resource,struct,socket
from cStringIO import StringIO
import version

SG_VER_REL=version.VERSION+'-'+version.RELEASE
BUFSIZE=10240
allowed_chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._-'

False=0
True=1

#########################################################################
### Debuging support

class debug_class:
  '''
  Debug class. Used for debugging informations.
  '''
  debug_level=2
  trace_level=9
  chroot=''
  def __init__(self):
      self.set_logfile('-')
  def dup(self):
      if self.fd>2:
        os.dup2(self.fd,1)
        os.dup2(self.fd,2)
  def set_logfile(self,logfile):
      self.logfile=logfile
      if logfile!='-':
        try:
          self.fd=os.open(logfile,os.O_CREAT|os.O_WRONLY|os.O_APPEND,0640)
        except OSError,(ec,es):
          debug.echo(0,"Error, can't open logfile: ",logfile,": ",es) 
        self.logfd=os.fdopen(self.fd,'a',0)
        #os.chmod(logfile,0640)
      else:
        self.logfd=os.fdopen(os.dup(1),'w',0)
        self.fd=1
      sys.stdout=self.logfd
      sys.stderr=self.logfd
  def reopen(self):
      if (self.logfile!='-') and (self.logfile[0:len(self.chroot)]==self.chroot):
        lfn=self.logfile[len(self.chroot):]
        if lfn[0]!="/":
          lfn="/"+lfn
        try:
          fd=os.open(lfn,os.O_CREAT|os.O_WRONLY|os.O_APPEND,0640)
          self.logfd=os.fdopen(self.fd,'a',0)
          os.dup2(fd,self.fd)
          os.close(fd)
          self.dup()
        except:
          debug.echo(0,"Warning, log is not reopened! (permission denied?)")
      else:
        debug.echo(0,"Error: logrotation not possible, log isn't in chroot")
  def set_level(self,level):
      self.debug_level=level
  def echo(self,level,*s):
      if self.debug_level>=level:
        o=''
        for i in s:
          if type(i)==type([]):
            for x in i:
              o+=str(x)
          else:
            o+=str(i)
        p=os.getpid()
        while 1:
          try:
            self.logfd.write("%5d: %s\n" % (p,o))
            break
          except IOError,(ec,es):
            if ec!=4:
              break
          except UnicodeEncodeError:
            o=re.sub('[^\x00-\x7F]','_',o) # leave only ascii chars
            break
  def stack(self,level,arg):
      self.echo(level,arg,traceback.format_stack())
  def traceback(self,level,arg):
      self.echo(level,arg,traceback.format_exception(sys.exc_type,
                       sys.exc_value, sys.exc_traceback))
  def traceback_value_str(self):
      return string.join(traceback.format_exception_only(
                             sys.exc_type, sys.exc_value),', ')

debug=debug_class()

#########################################################################
### Useful functions

def is_infected(level,detected=''):
    if level>=1.0:
      return True
    else:
      return False

class safe_class:
  ROOT_PATH='/'
  def fn(self,f):
      if f[0]=='/':
        # f.lstrip('/')
        while f[0]=='/':
          f=f[1:]
        return os.path.join(self.ROOT_PATH,f)
      else:
        return f
  def open(self,name,mode='',buffering=-1):
      if buffering>=0:
        return open(self.fn(name),mode,buffering)
      if mode:
        return open(self.fn(name),mode)
      return open(self.fn(name))
  def osopen(self,filename,flag,mode=0777):
      return os.open(self.fn(filename),flag,mode)

safe=safe_class()

def normalize_filename(sfn):
    '''replace !allowed_chars from filename with _'''
    ofn=''
    for chp in range(len(sfn)):
      if not sfn[chp] in allowed_chars:
        ofn=ofn+'_'
      else:
        ofn=ofn+sfn[chp]
    return ofn

def quote(string):
    return re.sub("([\\\\'])","\\\\\\1",string)

class popen:
  '''
  POpen implementation (with some changes)
  '''
  def __init__(self, cmd, wd='', resource_limits={}):
      self.exitstatus,self.killsignal,self.execerror,self.pid=-1,-1,'',0
      self.pcr,self.pcw = os.pipe()
      self.cpr,self.cpw = os.pipe()
      self.oldsigchld=signal.getsignal(signal.SIGCHLD)
      signal.signal(signal.SIGUSR1,self.sighandler)
      signal.signal(signal.SIGCHLD,self.sighandler)
      self.pid = os.fork()
      if self.pid == 0: # child
        signal.signal(signal.SIGUSR1,signal.SIG_DFL)
        signal.signal(signal.SIGCHLD,signal.SIG_DFL)
        os.close(self.pcw)
        os.close(self.cpr)
        os.dup2(self.pcr,0); os.close(self.pcr)
        os.dup2(self.cpw,1)
        os.dup2(self.cpw,2); os.close(self.cpw)
        try:
          if wd:
            os.chdir(wd)
          for key,value in resource_limits.items():
            resource.setrlimit(key,(value,value))
          debug.echo(4,"POpen: Running: ",str(cmd))
          os.execvp(cmd[0],cmd)
        except os.error,(ec,es):
          debug.echo(3,"POpen: os.error: ",es+str([cmd[0]]))
          os.write(1,es+str([cmd[0]]))
        os._exit(10)
      os.close(self.pcr)
      os.close(self.cpw)
      self.tocmd=os.fdopen(self.pcw, 'w', 0)
      self.fromcmd=os.fdopen(self.cpr, 'r', 0)
  def sighandler(self,sn,stack):
      if sn==signal.SIGUSR1:
        self.execerror=os.read(self.cpr,1024)
      else:
        self.wait()
        debug.echo(4,"POpen SIG=",sn,", pid=",self.pid,
              ", exitstatus=",self.exitstatus,
              ", signal=",self.killsignal)
  def wait(self):
      try:
        pid,es=os.waitpid(self.pid,0)
        self.exitstatus=es/256
        self.killsignal=es%256
      except OSError:
        if self.exitstatus<0:
          self.exitstatus=0 # pid already exited, but no exit status
      signal.signal(signal.SIGCHLD,self.oldsigchld)
      return self.exitstatus
  def readlines(self):
      try:
        return self.fromcmd.readlines()
      except IOError,(ec,es):
        if ec==4: #Interrupted system call
          return []
        else:
          raise
  def readline(self):
      try:
        return self.fromcmd.readline()
      except IOError,(ec,es):
        if ec==4: #Interrupted system call
          return []
        else:
          raise
  def write(self,s):
      if self.exitstatus<0:
        return self.tocmd.write(s)
      else:
        return 0
  def close(self,rw=3):
      if rw&1:
        self.tocmd.flush()
        os.close(self.pcw)
      if rw&2:
        os.close(self.cpr)

def rlistdir(rootpath,path=''):
    '''
    list files and directories recursively
    
    return values: [filelist],size
      [filelist] is an array of file and directory names
      size is added filesizes
    '''
    # decode to plain string from unicode
    rootpath=str(rootpath)
    path=str(path)
    files,fsize=[],0L
    try:
      listdir=os.listdir(safe.fn(os.path.join(rootpath,path)))
    except:
      return [],0L
    for d in listdir:
      stat=os.lstat(safe.fn(os.path.join(rootpath,path,d)))
      if stat[0] & 0x4000: # is_dir
        rld=rlistdir(rootpath,os.path.join(path,d))
        files=files+rld[0]
        fsize=fsize+rld[1]
      else: # is_file or is_symlink or ...
        fsize=fsize+stat[6]
        files.append(os.path.join(path,d))
    return files,fsize

def randomchars(count=10,chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    '''
    Return count random characters. Characters are randomly selected
    from chars array.
    '''
    s=''
    r=random.Random()
    for _ in range(count):
      s=s+chars[int(r.random()*len(chars))]
    return s

class mktemp:
      '''
      make a temporary file or directory
      Makeing of file:
        mktemp('filename-prefix','.suffix','w',0600).autorm()
      Makeinf od directory:
        mktemp('filename-prefix','.suffix','d',0600).autorm()
      '''
      def __init__(self,prefix,suffix='',flags='w',umode=-1):
          self.flag_autorm=0
          self.flags=flags
          while 1:
            s=randomchars()
            self.name=prefix+s+suffix
            self.name_no_prefix=s+suffix
            self.root_name=safe.fn(self.name)
            debug.echo(7,"mktemp: ",self.root_name)
            try:
              if flags=='d':
                if umode<0:
                  umode=0700
                os.mkdir(self.root_name)
                os.chmod(self.root_name,umode)
                if (os.stat(self.root_name)[0] & 0170000)==040000:
                  return
              else:
                if umode<0:
                  umode=0600
                self.fd=os.open(self.root_name,os.O_WRONLY|os.O_CREAT|os.O_EXCL,umode)
                self.f=os.fdopen(self.fd,'w',BUFSIZE)
                if (os.stat(self.root_name)[0] & 0170000)==0100000:
                  return
            except OSError,(ec,es):
              if not (ec in [17,21]):
                if ec in [2,13]:
                  debug.echo(1,"ERROR: Can't make temp file: ",self.root_name,": ",es)
                raise
      def __del__(self):
          if self.flag_autorm:
            self.rm()
      def rm(self):
          if self.flags=='d':
            os.rmdir(self.name)
          else:
            try:
              self.f.close()
            except:
              pass
            os.unlink(self.root_name)
      def autorm(self):
          self.flag_autorm=1
          return self

def tempcleanup(filename,s=''):
    '''
    Safe cleanup for a file.
    '''
    try:
      os.unlink(safe.fn(filename))
    except OSError,(er,es):
      debug.echo(0,s+": ERROR: OSError: ",es,filename)

def rdel(paths):
    '''
    deletes directory recursively
      paths is an array of directories or files
    '''
    for dirname in paths:
      try:
        if os.lstat(safe.fn(dirname))[0] & 0x4000: # is_dir
          for fname in os.listdir(safe.fn(dirname)):
            rdel([os.path.join(dirname,fname)])
          os.rmdir(safe.fn(dirname))
        else: # is fname or symlink
          os.unlink(safe.fn(dirname))
      except OSError,(ec,es):
        if not ec in [2]:
          debug.echo(3,"rdel: ",debug.traceback_value_str())

def fromhdr():
    '''
    Return a "From local@... DATE" string. String is by RFC.
    '''
    return "From local@local.local"+\
      time.strftime("  %c\r\n",time.localtime(time.time()))

def socket_settimeout(sock,seconds):
    '''
    Set a timeout for a socket. Timeout is in seconds.
    '''
    try:
      # try to use settimeout
      sock.settimeout(seconds)
    except AttributeError:
      # use this for python < 2.3
      sock.setsockopt(socket.SOL_SOCKET,socket.SO_RCVTIMEO,
        struct.pack('ll',seconds,0))

#########################################################################
### SMTP and mail classes

S_OK='SENT'
S_TEMPFAIL='TEMP-FAILED'
S_REJECT='REJECTED'
S_DROP='DROPPED'
S_FORCE_SEND='SEND-FORCED'

RECV_SMTP=0
RECV_HEADER=1 # obsolete
RECV_BODY=2
RECV_QUIT=3 # waiting for quit

class mail_class:
  '''
  EMail structure with some functions.
  '''
  reg_header=re.compile('^[!-9;-~]+:').search
  reg_header_cont=re.compile('^[ \t]+').search
  def __init__(self):
      self.addr='' # Connection address
      self.comm='' # SMTP communication
      self.header='' # Email Header
      self.data='' # Email Data (with header too)
      self.bodypos=0 # Body position
      self.xheader='' # Extended header
      self.xhdra={} # Extended header array
      self.recip=[] # recipient emails
      self.sender='' # sender
      self.df=StringIO()
  def close(self):
      self.df.flush()
      mail.data=self.df.getvalue()
      # check for header
      self.df.seek(0)
      lineno=0
      mail.header=''
      while True:
        lineno+=1
        l=self.df.readline()
        if (not l) | (l=='\r\n') | (l=='\n'):
          break
        mail.header+=l
        if self.reg_header(l):
          continue
        if lineno>1:
          if self.reg_header_cont(l):
            continue
      self.df.close()
      self.bodypos=len(mail.header)
  def normalize_header(self,lines,value=None):
      ''' normalize to <80 chars per line '''
      if value!=None:
        lines+=': '+value+'\r\n'
      SPACE8 = ' '*8
      MAXLINELEN2=78
      MAXLINELEN=MAXLINELEN2-len(SPACE8)
      sublines_semi=[]
      for line in lines.splitlines():
        if len(line.replace("\t", SPACE8)) <= MAXLINELEN2:
          sublines_semi.append(line.rstrip())
        else:
          # split line by semicolons
          while len(line) > MAXLINELEN:
            i=line.rfind(';', 0, MAXLINELEN)
            if i<0:
              break
            sublines_semi.append(line[:i+1].lstrip())
            line=line[i+1:]
          sublines_semi.append(line.lstrip())
      # now split sublines by spaces
      sublines_space=[]
      for line in sublines_semi:
        if len(line.replace("\t", SPACE8)) <= MAXLINELEN2:
          sublines_space.append(line)
        else:
          while len(line) > MAXLINELEN:
            i=line.rfind(' ', 0, MAXLINELEN)
            if i<0:
              break
            sublines_space.append(line[:i])
            line=line[i+1:]
          sublines_space.append(line)
      # and now join it
      sublines=''
      current=sublines_space.pop(0)
      while sublines_space:
        line=sublines_space.pop(0)
        nline=current+' '+line
        if len(nline.replace("\t", SPACE8)) <= MAXLINELEN2:
          current=nline
        else:
          sublines=sublines+current+"\r\n"
          current="\t"+line
      sublines=sublines+current+"\r\n"
      return sublines
  def addheader(self,desc,value=None):
      if value==None:
        hdr=desc
      else:
        hdr=desc+': '+value+'\r\n'
        self.xhdra[desc]=value
      self.xheader+=self.normalize_header(hdr)
  def modheader(self,frm,to):
      '''
      Modifies header by regular expression.
      Returns number of modifications (max. 1).
      '''
      h=re.compile(frm,re.MULTILINE).subn(to,self.header,1)
      if h[1]>0:
        self.header=h[0]
      return h[1]
  def hdrlen(self):
      return len(self.header)
  def addvirusinfo(self,detected,scannames=''):
      if not "X-Sagator-Scanner" in self.xhdra.keys():
        self.addheader("X-Sagator-Scanner",
          SG_VER_REL+' at '+socket.gethostname()+'; '+scannames)
      if not "X-Sagator-ID" in self.xhdra.keys():
        if globals.id:
          self.addheader("X-Sagator-id", globals.id)
      if detected:
        if not "X-Sagator-Status" in self.xhdra.keys():
          mail.addheader("X-Sagator-Status",
                         "%s [%s]" % (detected,globals.found_by.name))
  def getnamebyaddr(self,ip):
      try:
        return socket.gethostbyaddr(ip)[0]
      except:
        return ''

# mail object
mail=mail_class()

class SmtpcError(Exception):
  """SMTP Client Error"""
  ConnectError='ConnectError'
  WrongReturnCode='WrongReturnCode'
  SendmailError='SendmailError'

class smtp:
  '''
  Basic SMTP class. All SMTP clases are descended from this class.
  '''
  f=False
  bufsize=BUFSIZE
  SMTP_SERVER=('127.0.0.1',25)

class smtpc(smtp):
  '''
  SMTP client class.
  '''
  def __init__(self):
    try:
      self.conn=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(self.conn,120)
      self.conn.connect(self.SMTP_SERVER)
    except socket.error,(ec,es):
      debug.echo(0,"SMTPC: ERROR: connect: ",es)
      raise SmtpcError,(SmtpcError.ConnectError,"451 SMTP connection refused\r\n")
    self.f=self.conn.makefile('rw',self.bufsize)
  def __del__(self):
    self.close()
  def send(self,data,rc='...'):
    if data!='':
      debug.echo(5,"SMTPC>: ",data.rstrip())
      self.conn.sendall(data+"\r\n")
    ret=''
    while 1:
      r=self.readline()
      ret=ret+r
      if re.search("^[0-9][0-9][0-9] ",r):
        break
    debug.echo(5,"SMTPC<: ",ret.rstrip())
    if not re.search("^"+rc+" ",ret):
      raise SmtpcError,(SmtpcError.WrongReturnCode,ret.rstrip()+"!="+rc)
    return ret
  def readline(self):
    return self.f.readline()
  def close(self):
    try:
      self.f.close()
      self.conn.shutdown(2)
      self.conn.close()
    except:
      pass
  def sendmail(self,sender,recipients,ldata=''):
    ret="451 SMTP connection error\r\n"
    try:
      self.send('','220')
      self.send("HELO "+socket.gethostname(),'250')
      self.send("MAIL FROM:"+sender,'250')
      for rec in recipients:
        self.send("RCPT TO:"+rec,'250')
      self.send("DATA",'354')
      self.conn.sendall(ldata)
      debug.echo(5,"SMTPC>: DATA SENT")
      ret=self.send(".","250")
    except SmtpcError.WrongReturnCode,ret:
      raise SmtpcError,(SmtpcError.SendmailError,ret)
    try:
      self.send("QUIT","221")
      self.close()
    except SmtpcError.WrongReturnCode,ret:
      debug.echo(5,"SMTPC: quit error",ret)
    return ret

#########################################################################
### Scanner templates

class ScannerError(Exception):
      """Any Error Of Any Scanner"""

class scanoper:
  '''
  A scanner used to operate over various scanners.
  '''
  name='op()'
  is_scanner=True
  is_multistream=False
  is_interscanner=True
  ignore_name=True
  filename='stream'
  def __init__(self,op,a,b=None):
      self._oper=(op,a,b)
      if b:
        self.name=a.name+op+b.name
      else:
        self.name=a.name
      self.scanner=a
  def _eval(self,l1,d1,v1,l2,d2,v2):
      op=self._oper[0]
      if op=='+':
        return l1+l2,d1+','+d2,v1+v2
      elif op=='-':
        return l1-l2,d1+','+d2,v1+v2
      elif op=='*':
        return l1*l2,d1+','+d2,v1+v2
      elif op=='/':
        return l1/l2,d1+','+d2,v1+v2
      else:
        raise ScannerError,"Unknown operator: "+self._oper[0]
  def scanstream(self,stream):
      l,d,v={},{},{}
      if self._oper[0] in '+-*/':
        for i in [1,2]:
          self._oper[i].prescan()
          l[i],d[i],v[i]=self._oper[i].scanstream(stream)
          self._oper[i].postscan(l[i],d[i],v[i])
        return self._eval(l[1],d[1],v[1],l[2],d[2],v[2])
      elif self._oper[0]=='|':
        try:
          self._oper[1].prescan()
          l,d,v=self._oper[1].scanstream(stream)
          self._oper[1].postscan(l,d,v)
          return l,d,v
        except:
          self._oper[2].prescan()
          l,d,v=self._oper[2].scanstream(stream)
          self._oper[2].postscan(l,d,v)
          return l,d,v
      elif self._oper[0]=='&':
        self._oper[1].prescan()
        l1,d1,v1=self._oper[1].scanstream(stream)
        self._oper[1].postscan(l1,d1,v1)
        if is_infected(l1):
          self._oper[2].prescan()
          l2,d2,v2=self._oper[2].scanstream(stream)
          self._oper[2].postscan(l2,d2,v2)
          return l1*l2,d1+'*'+d2,v1+v2
        else:
          return 0.0,'',[]
      elif self._oper[0]=='~':
        try:
          self._oper[1].prescan()
          l,d,v=self._oper[1].scanstream(stream)
        except:
          debug.echo(4,"Scanner %s failed, recovering..." % self._oper[1].name)
          if self._oper[1].is_interscanner:
            l,d,v=self._oper[1].get('child_status')
          else:
            return 0.0,'',[]
        self._oper[1].postscan(l,d,v)
        return l,d,v
      else:
        raise ScannerError,"Unknown operator: "+self._oper[0]
  def scanfile(self,files,dirname=''):
      l,d,v={},{},{}
      if self._oper[0] in '+-*/':
        for i in [1,2]:
          self._oper[i].prescan()
          l[i],d[i],v[i]=self._oper[i].scanfile(files,dirname)
          self._oper[i].postscan(l[i],d[i],v[i])
        return self._eval(l[1],d[1],v[1],l[2],d[2],v[2])
      elif self._oper[0]=='|':
        try:
          self._oper[1].prescan()
          l,d,v=self._oper[1].scanfile(files,dirname)
          self._oper[1].postscan(l,d,v)
          return l,d,v
        except:
          self._oper[2].prescan()
          l,d,v=self._oper[2].scanfile(files,dirname)
          self._oper[2].postscan(l,d,v)
          return l,d,v
      elif self._oper[0]=='&':
        self._oper[1].prescan()
        l1,d1,v1=self._oper[1].scanfile(files,dirname)
        self._oper[1].postscan(l1,d1,v1)
        if is_infected(l1):
          self._oper[2].prescan()
          l2,d2,v2=self._oper[2].scanfile(files,dirname)
          self._oper[2].postscan(l2,d2,v2)
          return l1*l2,d1+'*'+d2,v1+v2
        else:
          return 0.0,'',[]
      elif self._oper[0]=='~':
        try:
          self._oper[1].prescan()
          l,d,v=self._oper[1].scanfile(files,dirname)
        except:
          debug.echo(4,"Scanner %s failed, recovering..." % self._oper[1].name)
          if self._oper[1].is_interscanner:
            l,d,v=self._oper[1].get('child_status')
          else:
            return 0.0,'',[]
        self._oper[1].postscan(l,d,v)
        return l,d,v
      else:
        raise ScannerError,"Unknown operator: "+self._oper[0]
  def prescan(self):
      pass
  def postscan(self,level,vir,ret):
      pass
  def destroy(self):
      self._oper[1].destroy()
      if self._oper[2]:
        self._oper[2].destroy()
  def reinit(self):
      self._oper[1].reinit()
      if self._oper[2]:
        self._oper[2].reinit()
  def get(self,var):
      try:
        return eval("self._oper[1]."+var)
      except AttributeError:
        try:
          return eval("self._oper[2]."+var)
        except AttributeError:
          return "UNKNOWN"
  def __add__(self,second):
      return scanoper('+',self,second)
  def __sub__(self,second):
      return scanoper('-',self,second)
  def __mul__(self,second):
      return scanoper('*',self,second)
  def __div__(self,second):
      return scanoper('/',self,second)
  def __or__(self,second):
      return scanoper('|',self,second)
  def __and__(self,second):
      return scanoper('&',self,second)
  def __invert__(self):
      return scanoper('~',self)

class ascanner(scanoper):
  '''
  Default scanner user for building all other realscanners.
  '''
  name='AScanner()'
  is_spamscan=False
  is_interscanner=False
  ignore_name=False
  scanner=0
  def __init__(self):
      pass
  def destroy(self):
      pass
  def prescan(self):
      '''This function is called before running a scanner'''
      debug.echo(4,"Running: ",self.name)
  def postscan(self,level,vir,ret):
      '''This function is called after running a scanner'''
      self.child_status=(level,vir,ret)
      if is_infected(level):
        if not self.ignore_name:
          globals.found_by=self
          debug.echo(5,"Found %s by %s, level %f" % (vir,self.name,level))
  def scanstream(self,stream):
      raise ScannerError,'Not implemented'
  def scanfile(self,files,dirname=''):
      raise ScannerError,'Not implemented'
  def reinit(self):
      pass
  def get(self,var):
      try:
        return eval("self."+var)
      except AttributeError:
        return "UNKNOWN"

class interscanner(ascanner):
  '''
  Default scanner used for building all other interscanners.
  '''
  name='AInterScanner()'
  is_interscanner=True
  scanners=[]
  def rename(self,scanners):
      n=[]
      for s in scanners:
        n.append(s.name)
      self.name=self.name.replace("()","("+string.join(n,', ')+")")
      if self.name=="":
        self.name=string.join(n,',')
  def destroy(self):
      self.scanner.destroy()
  def prescan(self):
      '''This function is called before running a scanner'''
      debug.echo(5,"Running: ",self.name)
  def postscan(self,level,vir,ret):
      '''This function is called after running a scanner'''
      self.child_status=(level,vir,ret)
  def reinit(self):
      self.scanner.reinit()
  def get(self,var):
      try:
        r=eval("self."+var)
        if not r:
          r=self.scanner.get(var)
        return r
      except AttributeError:
        return self.scanner.get(var)

class globals_class:
  ACTION=S_REJECT
  QFNAME='' # quarantine filename (generated)
  USER=''
  GROUP=''
  CHROOT=''
  id=None # sagator's message ID
  scan_only=False # only scan or also send reports?
  daemon=True # daemonize or not?
  service_pids=[] # service PIDs
  pid_file=None # PID filename
  pidf=None # PID file (as python file object)
  def __init__(self):
      self.reset()
  def reset(self):
      self.found_by=ascanner
      self.QFNAME=''
  def action(self,level,detected=''):
      if is_infected(level):
        return self.ACTION
      else:
        return S_OK
globals=globals_class()
